# -*- coding: utf-8 -*-
#
# ApplicationIntegrationServer - Provides an extensible Application 
# Integration Server for MailStore
#
# Copyright (c) 2014, 2015 MailStore Software GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import ssl
import sys
import json
import base64
import logging
import logging.handlers
import importlib
import http.server
import urllib.parse
import socketserver
import configparser


class RequestHandler(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        """Handle HTTP GET requests by parsing the query string parameters"""
        raw = urllib.parse.urlparse(self.path).query
        self.server.log.debug("{}:{} GET data: {}".format(self.client_address[0], self.client_address[1], raw))
        data = urllib.parse.parse_qs(raw)
        self.handle_request(data)

    def do_POST(self):
        """Handle HTTP POST requests by parsing the POST body"""
        raw_body = self.rfile.read(int(self.headers["Content-Length"])).decode()
        self.server.log.debug("{}:{} POST data: {}".format(self.client_address[0], self.client_address[1], raw_body))
        data = urllib.parse.parse_qs(raw_body)
        # Also handle query parameter
        raw_query = urllib.parse.urlparse(self.path).query
        self.server.log.debug("{}:{} QUERY data: {}".format(self.client_address[0], self.client_address[1], raw_query))
        data.update(urllib.parse.parse_qs(raw_query))
        self.handle_request(data)

    def log_message(self, format_, *args):
        """This method overwrites the default method to suppress writing to STDERR,
        which fails when running as a service, making the service completely unusable.
        See http://stackoverflow.com/questions/10651052/how-to-quiet-simplehttpserver for details."""
        self.server.log.debug(
            ("%s - - [%s] %s\n" % (self.address_string(), self.log_date_time_string(), format_ % args)).rstrip())

    def is_http_user_authenticated(self, plugin):
        """If a plugin requires additional authentication, try to authenticate the user."""

        # No authentication required by plugin
        if not hasattr(self.server.plugins[plugin], "http_authentication"):
            self.server.log.debug("{}:{} NO authentication required".format(self.client_address[0], self.client_address[1]))
            return True

        # Authentication required by plugin
        self.server.log.debug("{}:{} authentication required".format(self.client_address[0], self.client_address[1]))
        authorize_header = self.headers["authorization"]

        # Authorization required, but no header is sent by the client
        if not authorize_header:
            self.server.log.debug("{}:{} NO auth header sent by client".format(self.client_address[0], self.client_address[1]))
            return False

        self.server.log.debug("{}:{} auth header: {}".format(self.client_address[0], self.client_address[1], authorize_header))
        type_, value = authorize_header.split()
        decoded_value = base64.b64decode(value.encode()).decode()
        username, password = decoded_value.split(":")

        # Send http username and password to plugin
        is_authenticated = self.server.plugins[plugin].http_authentication(username, password)
        self.server.log.debug(
            "{}:{} credentials: '{}' '{}', is authenticated: {}".format(self.client_address[0],
                                                                        self.client_address[1], username,
                                                                        password, is_authenticated))
        return is_authenticated

    def handle_request(self, data):
        """Handle parsed request data from do_GET and do_POST methods.
        
        data: Dictionary of request parameters
        """

        self.server.log.debug("{}:{} parsed data: {}".format(self.client_address[0], self.client_address[1], data))
        path = urllib.parse.urlparse(self.path).path
        self.server.log.debug("{}:{} path: {}".format(self.client_address[0], self.client_address[1], path))
        path = path.strip("/")

        # Path must match a plugin name, else return HTTP 404
        if path not in self.server.plugins.keys():
            self.server.log.warning("{}:{} HTTP 404 Plugin '{}' not found".format(self.client_address[0], self.client_address[1], path))
            self.send_404(path)
            return

        # Check if http user is authenticated, else return HTTP 401
        if not self.is_http_user_authenticated(path):
            self.server.log.warning("{}:{} HTTP 401 authentication required".format(self.client_address[0], self.client_address[1]))
            self.send_401()
            return

        # Parameter 'cmd' must always be present, else return HTTP 400
        if not "cmd" in data:
            message = "'cmd' parameter missing."
            self.server.log.warning("{}:{} HTTP 400 {}".format(self.client_address[0], self.client_address[1], message))
            self.send_400(message)
            return

        # List users
        if data["cmd"][0] == "list":
            try:
                # Execute list_() method from plugin that matches path name
                user_list = self.server.plugins[path].list_(data)
            except Exception as e:
                self.server.log.error("{}:{} HTTP 500 user list of '{}': {}".format(self.client_address[0],
                                                                                    self.client_address[1],
                                                                                    path, str(e)))
                # Don't send 'e' to the client, because sensitive information could leak
                self.send_500()
            else:
                self.server.log.info("{}:{} HTTP 200 user list of '{}'".format(self.client_address[0],
                                                                               self.client_address[1],
                                                                               path))
                try:
                    self.send_200(user_list)
                except Exception as e:
                    self.server.log.error("SERVER ERROR while sending HTTP 200 response to list_: {}".format(repr(e)))
                else:
                    self.server.log.info("SERVER HTTP 200 OK")
        # Authenticate a user
        elif data["cmd"][0] == "auth":
            # Parameters 'user' and 'pass' must be given, else return HTTP 400
            if not ("user" in data and "pass" in data):
                message = "'user' or 'pass' parameter missing."
                self.server.log.warning("{}:{} HTTP 400 {}".format(self.client_address[0], self.client_address[1], message))
                self.send_400(message)
                return

            # Execute auth() method from plugin that matches path name
            try:
                user_is_authenticated = self.server.plugins[path].auth(data["user"][0], data["pass"][0])
            except Exception as e:
                self.server.log.error("{}:{} HTTP 500 auth of '{}': {}".format(self.client_address[0],
                                                                               self.client_address[1],
                                                                               data["user"][0],
                                                                               repr(e)))
                # Don't send 'e' to the client, because sensitive information could leak
                self.send_500()
            else:
                if isinstance(user_is_authenticated, bool):
                    self.server.log.warning(
                        "PLUGIN '{}' returns 'bool' type instead of a 'dict' type to an 'auth' call."
                        "performing conversion".format(path))
                    user_is_authenticated = {"succeeded": user_is_authenticated}
                self.server.log.info(
                    "{}:{} HTTP 200 auth: {}@{} {}".format(self.client_address[0],
                                                           self.client_address[1],
                                                           data["user"][0],
                                                           path, str(user_is_authenticated)))
                try:
                    self.send_200(user_is_authenticated)
                except Exception as e:
                    self.server.log.error("SERVER ERROR while sending HTTP 200 response to auth: {}".format(repr(e)))
                else:
                    self.server.log.info("SERVER HTTP 200 OK")
        # No valid cmd parameter value
        else:
            message = "unknown command: {}.".format(data["cmd"][0])
            self.server.log.warning(
                "{}:{} HTTP 400 {}".format(self.client_address[0], self.client_address[1], message))
            self.send_400(message)

    def send_200(self, data):
        """Send HTTP 200 OK response

        data: Response body
        """
        data = json.dumps(data).encode()
        self.send_response(200)
        self.send_header("Content-Type", "application/json; charset=utf-8")
        self.send_header("Content-Length", len(data))
        self.end_headers()
        self.server.log.debug("SERVER sending data: {}".format(data.decode()))
        self.wfile.write(data)

    def send_400(self, message=""):
        """Send HTTP 400 Bad request"""
        error_message = "<!DOCTYPE html><head><title>MailStore Application Integration Server</title></head>\
                        <body>400: Bad request.<br/>{0}</body></html>".format(message).encode()
        self.send_response(400)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", len(error_message))
        self.end_headers()
        self.wfile.write(error_message)

    def send_401(self):
        """Send HTTP 401 Authentication required"""
        error_message = "<!DOCTYPE html><head><title>MailStore Application Integration Server</title></head>\
                        <body>401: Authentication required.</body></html>".encode()
        self.send_response(401)
        self.send_header("www-authenticate", 'Basic Realm=""')
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", len(error_message))
        self.end_headers()
        self.wfile.write(error_message)

    def send_404(self, path):
        """Send HTTP 404 Plugin not found
        
        path: Path representing the plugin name
        """
        error_message = "<!DOCTYPE html><head><title>MailStore Application Integration Server</title></head>\
                        <body>404: Plugin '{}' not found.</body></html>".format(path).encode()
        self.send_response(404)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", len(error_message))
        self.end_headers()
        self.wfile.write(error_message)

    def send_500(self):
        """Send HTTP 500 Internal Server Error"""
        error_message = "<!DOCTYPE html><head><title>MailStore Application Integration Server</title></head>\
                        <body>500: Internal Server Error</body></html>".encode()
        self.send_response(500)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", len(error_message))
        self.end_headers()
        self.wfile.write(error_message)


class Server(socketserver.ThreadingMixIn, http.server.HTTPServer):
    def __init__(self, *args, **kwargs):
        # Initialize HTTP server
        self.configuration = kwargs["configuration"]

        self.log = self.init_logging()

        self.log.warning("Starting Application Integration Server")

        # Load plugins
        self.log.info("SERVER loading plugins")
        self.plugins = self.init_plugins()
        self.log.info("SERVER plugins used: {}".format(", ".join(self.plugins.keys())))

        # Remove "configuration" so **kwargs can be passed to http.server.HTTPServer.__init__()
        del kwargs["configuration"]

        # Start HTTP server
        http.server.HTTPServer.__init__(self, *args, **kwargs)
        self.log.warning("SERVER listening on {}:{}".format(*self.server_address))
        
        # Will be set to false, in case the certificate cannot be found, then the server will not be started
        self.operational = True
        self.init_tls()

    def init_logging(self):
        """Initialize logging"""
        logger = logging.getLogger("AppIntLogger")
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(module)s.%(funcName)s() [%(lineno)d]: %(message)s")

        # Set up logging to standard out if it is enabled
        if self.configuration["log"].getboolean("stdout"):
            streamhandler = logging.StreamHandler(sys.stdout)
            config_stdout_level = self.configuration["log"]["stdout_level"]
            streamhandler.setLevel(getattr(logging, config_stdout_level))
            streamhandler.setFormatter(formatter)
            logger.addHandler(streamhandler)

        # Set up logging to file system if it is enabled
        if self.configuration["log"].getboolean("file"):
            directory = self.configuration["log"]["directory"]
            os.makedirs(directory, exist_ok=True)

            debug_log_file = os.path.join(directory, "AppInt.log")
            filehandler = logging.handlers.TimedRotatingFileHandler(debug_log_file, when="D", interval=1,
                                                                    backupCount=60)
            config_file_level = self.configuration["log"]["file_level"]
            filehandler.setLevel(getattr(logging, config_file_level))
            filehandler.setFormatter(formatter)
            logger.addHandler(filehandler)

        return logger

    def init_plugins(self, plugin_dir="plugins"):
        """Initialize plugins"""
        plugin_dir_absolute = os.path.join(os.path.dirname(os.path.abspath(__file__)), plugin_dir)
        self.log.info("PLUGIN directory: {}".format(plugin_dir_absolute))

        # Load the plugin classes into a dictionary, where the key is the filename.
        # This allows us to call the plugin methods as follows:
        #     self.server.plugins[pluginname].list_() or .auth(username, password)
        # We can pass the "path" of the HTTP command as parameter to the 'server.plugins' dictionary
        plugins = dict()
        for file in os.listdir(plugin_dir_absolute):
            self.log.debug("PLUGIN {} file found".format(file))

            if not file.endswith(".py"):
                self.log.info("PLUGIN {} ignored, as it is not a python file".format(file))
                continue
            if file.startswith("_"):
                self.log.info("PLUGIN {} ignored, file does not meet naming convention".format(file))
                continue

            # Remove file extension
            plugin_file = os.path.splitext(file)[0]
            self.log.info("PLUGIN {} importing".format(plugin_file))
            try:
                # Import plugin file as module
                plugin_import_object = importlib.import_module("{}.{}".format(plugin_dir, plugin_file))
            except Exception as e:
                self.log.error("PLUGIN {} ERROR while importing: {}".format(plugin_file, repr(e)))
            else:
                self.log.debug("PLUGIN {} imported successfully, loading 'Plugin' class".format(plugin_file))
                try:
                    # Instantiate Plugin() class
                    plugin_class = plugin_import_object.Plugin()
                except Exception as e:
                    self.log.error("PLUGIN {}.Plugin() class cannot be loaded: {}".format(plugin_file, repr(e)))
                else:
                    self.log.debug("PLUGIN {} loaded successfully".format(plugin_file))

                    # Check if plugin has all necessary properties
                    if not hasattr(plugin_class, "auth"):
                        self.log.error("PLUGIN {} cannot be used: 'auth' method missing".format(plugin_file))
                        continue
                    if not hasattr(plugin_class, "list_"):
                        self.log.error("PLUGIN {} cannot be used: 'list_' method missing".format(plugin_file))
                        continue

                    # Check if plugin requires authentication
                    if hasattr(plugin_class, "http_authentication"):
                        self.log.info("PLUGIN {} requires HTTP authentication".format(plugin_file))

                    # Add plugin to global list
                    plugins[plugin_file] = plugin_class
                    self.log.info("PLUGIN {} imported successfully".format(plugin_file))

        return plugins

    def init_tls(self):
        """Initialize TLS encryption"""

        # Load certificate and initialize TLS only if it is enabled in the configuration
        if self.configuration["tls"].getboolean("enabled"):
            # Convert relative path to certificate file into absolute path...
            cert_file_absolute = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                              self.configuration["tls"]["certificate"])

            # ...otherwise assume absolute path was already given.
            if not os.path.exists(cert_file_absolute):
                cert_file_absolute = self.configuration["tls"]["certificate"]

            # When a certificate could be found, wrap the socket
            if os.path.exists(cert_file_absolute):
                self.log.info("SERVER TLS certificate: {}".format(cert_file_absolute))
                self.socket = ssl.wrap_socket(self.socket, certfile=cert_file_absolute,
                                              server_side=True, ssl_version=ssl.PROTOCOL_TLSv1_2)
                self.log.info("SERVER TLS activated")

            else:
                self.operational = False
                self.log.critical("SERVER TLS certificate {} not found, server will not be started".format(
                    cert_file_absolute))

        # TLS is disabled in the configuration
        else:
            self.log.warning("SERVER TLS disabled in global configuration")


def main():
    # Read configuration from 'config.ini' file
    configuration = configparser.ConfigParser()
    configuration.read(os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.ini"))
    ipaddress = configuration["network"]["ipaddress"]
    port = configuration["network"].getint("port")

    # Initialize the server
    server = Server((ipaddress, port), RequestHandler, configuration=configuration)

    # Start server if initialization was successful
    if server.operational:
        server.serve_forever()
    else:
        sys.exit(1)


if __name__ == '__main__':
    main()
