# -*- coding: utf-8 -*-
#
# ApplicationIntegrationServer - LDAP Example Plugin
#
# Copyright (c) 2014 - 2017 MailStore Software GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# https://pypi.python.org/pypi/ldap3
# https://ldap3.readthedocs.org/
# https://media.readthedocs.org/pdf/ldap3/latest/ldap3.pdf
import ldap3
import logging


class Plugin:
    def __init__(self):
        self.ldap_host = "dc"
        self.ldap_port = 389  # ssl=636
        self.ldap_ssl = False
        self.ldap_netbios_domain = "example"
        self.ldap_user = "username"
        self.ldap_password = "password"
        self.ldap_base_dn = "OU=Users,DC=example,DC=com"
        self.ldap_searchfilter = "(objectClass=user)"
        self.ldap_server = ldap3.Server(self.ldap_host, port=self.ldap_port, use_ssl=self.ldap_ssl, get_info=ldap3.ALL)
        self.logger = logging.getLogger("AppIntLogger")
        # credentials to use this plugin
        self.plugin_user = "admin"
        self.plugin_password = "password"

    def list_(self, _):
        users = list()
        connection = ldap3.Connection(self.ldap_server,
                                      user="{}\\{}".format(self.ldap_netbios_domain, self.ldap_user),
                                      password=self.ldap_password,
                                      raise_exceptions=True)
        connection.bind()
        connection.search(self.ldap_base_dn,
                          self.ldap_searchfilter,
                          ldap3.SUBTREE,
                          attributes=['sAMAccountName', 'displayName', 'mailNickname', 'proxyAddresses'])
        # print(connection.result)
        for user in connection.response:
            temp_user = {}
            if 'attributes' in user.keys():
                if "sAMAccountName" in user['attributes'].keys() and user['attributes']["sAMAccountName"]:
                    temp_user["userName"] = user['attributes']["sAMAccountName"].lower()
                else:
                    self.logger.info("LDAP no sAMAccountName found, skipping user {}".format(repr(user)))
                    continue

                if "dn" in user.keys():
                    temp_user["distinguishedName"] = user["dn"]
                if "displayName" in user['attributes'].keys():
                    if user['attributes']["displayName"]:
                        temp_user["fullName"] = user['attributes']["displayName"]
                    else:
                        temp_user["fullName"] = ""
                if "proxyAddresses" in user['attributes'].keys():
                    try:
                        temp_user["emailAddresses"] = [email_address.split(":")[1].lower() for email_address
                                                   in user['attributes']["proxyAddresses"]]
                    except Exception as e:
                        self.logger.error(repr(e))
                users.append(temp_user)
            else:
                self.logger.warning("LDAP no attributes found, skipping user {}".format(repr(user)))
        connection.unbind()
        del connection
        return users

    def auth(self, username, password):
        if not username.startswith(self.ldap_netbios_domain):
            username = "{}\\{}".format(self.ldap_netbios_domain, username)
        connection = ldap3.Connection(self.ldap_server, user=username, password=password, raise_exceptions=False)
        connected = connection.bind()
        if not connected:
            connection.unbind()
        del connection
        return {"succeeded": connected}

    def http_authentication(self, username, password):
        return username == self.plugin_user and password == self.plugin_password
