# -*- coding: utf-8 -*-
#
# ApplicationIntegrationServer - AppInt Example Plugin
# This plugin connects to other Application Integration servers/plugins/urls.
# It aggregates users of all configures Application Integration servers.
# Authentication is tested against all of them. If one server sends an "OK",
# no other server will be queried.
#
# Copyright (c) 2014 - 2017 MailStore Software GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urllib.request
import urllib.parse
import base64
import json
import ssl


class Plugin:
    def __init__(self):
        # remote_servers must be a list of dictionaries with "user", "password" and "url" keys
        # set user and password to "" (empty string) to disable authentication
        # http and https are supported
        self.remote_servers = {"csv": {"credentials": "admin:password", "url": "https://localhost:2020/exCSVSimple"},
                               "simple": {"credentials": "admin:password", "url": "https://localhost:2020/exSimple"},
                               "ini": {"credentials": "admin:password", "url": "https://localhost:2020/exINI"}}
        self.cached_mode = True  # user_db will only be updated when a list is performed
        self.user_db = dict()
        # credentials to use this plugin
        self.plugin_user = "admin"
        self.plugin_password = "password"

    @staticmethod
    def _create_request(server, values):
        post_data = urllib.parse.urlencode(values).encode()
        request = urllib.request.Request(server["url"], post_data)
        if "credentials" in server.keys() and server["credentials"]:
            auth_header = base64.b64encode(server["credentials"].encode()).decode()
            request.add_header("Authorization", "Basic {}".format(auth_header))
        answer = urllib.request.urlopen(request, context=ssl._create_unverified_context()).read()
        answer = answer.decode()
        return json.loads(answer)

    def list_(self, _):
        users = list()
        for name, server in self.remote_servers.items():
            new_users = self._create_request(server, [("cmd", "list")])
            for new_user in new_users:
                if new_user["userName"] not in [user["userName"] for user in users]:
                    users.append(new_user)
                else:
                    raise Exception("duplicate user error {}".format(new_user["userName"]))
            if self.cached_mode:
                self.user_db[name] = new_users
        return users

    def auth(self, username, password):
        if self.cached_mode:
            for name, values in self.user_db.items():
                if username in [user["userName"] for user in values]:
                    return self._create_request(self.remote_servers[name],
                                                [("cmd", "auth"), ("user", username), ("pass", password)])
        else:
            for name, values in self.remote_servers.items():
                if username in [user["userName"] for user in self._create_request(values, [("cmd", "list")])]:
                    return self._create_request(self.remote_servers[name],
                                                [("cmd", "auth"), ("user", username), ("pass", password)])
        # fail, if no server lists 'username'
        return {"succeeded": False}

    def http_authentication(self, username, password):
        return username == self.plugin_user and password == self.plugin_password
