# -*- coding: utf-8 -*-
#
# ApplicationIntegrationServer - Office 365 Example plugin
#
# Copyright (c) 2014 - 2017 MailStore Software GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urllib.request
import urllib.parse
import urllib.error
import json
import base64


class Plugin:
    def __init__(self):
        self.tenantID = 'office365.mailstoretest.com'        # One of the domains registered for the tenant. Can be used nowadays instead of the tenant's ObjectId which wasn't possible in earlier Azure versions.
        self.clientID = 'MailStoreServer/mailstoretest.com'  # One of the ServicePrincipalNames of the service principal. Its AppPrincipalId is used as a default ServicePrincipalName.
        self.clientSecret = 'Passw0rd'                       # The password of the service principal.

        # Office 365 / Azure APIs are RESTful with defined endpoints for each tenant. Authentication can be done using OAUTH2.

        self.oauthTokenURI = 'https://login.windows.net/{0}/oauth2/token?api-version=1.0'.format(self.tenantID) # The tenant endpoint for obtaining an OAUTH2 token. API version is static (1.0) but needs to be supplied as a query parameter.
        # $oauthAuthURI = 'https://login.windows.net/{0}/oauth2/authorize?api-version=1.0&{1}'        # We don't need to use OAUTH2 authorisation in a user context but mention the endpoint for completeness.
        self.graphResourceURI = 'https://graph.windows.net'                                                # The Azure Graph API provides Azure AD access: http://msdn.microsoft.com/en-us/library/azure/hh974476.aspx
        self.graphAPIVersion = '1.5'                                                                       # The current Graph API version.
        self.graphAPIURI = self.graphResourceURI + '/{0}/{1}?api-version=' + self.graphAPIVersion                    # The Graph API Rest endpoint. API version is non-static (currently 1.5) and needs to be supplied as a query parameter.
        self.o365EWSURI = "https://outlook.office365.com/ews/exchange.asmx"
        # credentials to use this plugin
        self.plugin_user = "admin"
        self.plugin_password = "password"

    def list_(self, _):
        """Returns a list of dictionaries, in which the user parameters are defined"""
        """

          We need a JSON Web Token (JWT) to access the Graph API. To obtain the access token we need to send a POST request to the oauthTokenURI with these parameters in the body:

          grant_type:     Needs to be set "client_credentials" for access through a service principal.
          resource:       The base URI of the API we want to access, here "https://graph.windows.net".
          client_id:      One of the ServicePrincipalNames of the service principal.
          client_secret:  The password of the service principal.

        """

        # We store the parameters in a dictionary.

        parameters = {'grant_type': 'client_credentials',
                      'resource': self.graphResourceURI,
                      'client_id': self.clientID,
                      'client_secret': self.clientSecret}
        postData = urllib.parse.urlencode(parameters)
        postData = postData.encode('utf-8')

        """

          If the request below succeeds, we get an access token like this:

          token_type   : Always 'Bearer'
          access_token : The Base64 encrypted JWT.
          expires_in   : Expiry period in seconds, defaults to 43199 (12h minus 1 second)
          not_before   : A JSON numeric value representing the number of seconds from 1970-01-01T00:00:00Z UTC until the specified UTC date/time, ignoring leap seconds.
          expires_on   : A JSON numeric value representing the number of seconds from 1970-01-01T00:00:00Z UTC until the specified UTC date/time, ignoring leap seconds.
          resource     : The resource to be accessed, here: 'https://graph.windows.net'

        """

        accessToken = json.loads(
            ((urllib.request.urlopen(urllib.request.Request(self.oauthTokenURI, postData))).read()).decode())

        # The JWT has to be supplied which each request to the Graph API as an Authorization header akin to BasicAuth in the form: 'Authorization: Bearer <JWT>'

        headers = {"Authorization": "{0} {1}".format(accessToken['token_type'], accessToken['access_token'])}

        # Now we can request some Azure AD info at last!

        # Get all users in our tenant
        o365users = json.loads(((urllib.request.urlopen(
            urllib.request.Request(self.graphAPIURI.format(self.tenantID, "users"), None, headers))).read()).decode())['value']

        # We don't need all attributes of the Office 365 users, just some
        users = []
        for o365user in o365users:
            users.append(
                {
                    "userName": o365user['userPrincipalName'],
                    "distinguishedName": o365user['userPrincipalName'],
                    "fullName": o365user['displayName'],
                    "emailAddresses": o365user['proxyAddresses']
                }
            )

        return users

    def auth(self, username, password):
        """Authenticate username against external database.
        Must return True or False"""

        authstring = ('%s:%s' % (username, password)).encode('utf-8')
        base64string = base64.b64encode(authstring)
        authheader =  {"Authorization": "Basic %s" % base64string.decode('ascii')}
        request = urllib.request.Request(self.o365EWSURI, None, authheader)
        try:
            handle = urllib.request.urlopen(request)
        except urllib.error.HTTPError as e:
            if e.code == 401:
                return {"succeeded": False}
            else:
                raise e
        return {"succeeded": True}

    def http_authentication(self, username, password):
        return username == self.plugin_user and password == self.plugin_password
