#!/usr/bin/env python3

# check_mailstorelicence.py - Checks the amount of available licences.
#
# Copyright (c) 2012 - 2018 MailStore Software GmbH
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicence, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import argparse
import mailstore

if __name__ == "__main__":

    # Initialize command line argument parser
    parser = argparse.ArgumentParser(description="Checks the amount of available licenses.", 
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("--username", "-u",
                        default="admin",
                        help="MailStore Username")

    parser.add_argument("--password", "-p",
                        default="admin",
                        help="MailStore Password")

    parser.add_argument("--host", "-H",
                        default="localhost",
                        help="MailStore Server Hostname/IP")

    parser.add_argument("--port", "-P",
                        type=int,
                        default=8463,
                        help="MailStore Server Port")

    parser.add_argument("--licensed", "-l",
                        type=int,
                        default=None,
                        required=True,
                        help="Number of licensed MailStore users")

    parser.add_argument("--warning", "-w",
                        type=int,
                        default=None,
                        required=True,
                        help="Warning Threshold")

    parser.add_argument("--critical", "-c",
                        type=int,
                        default=None,
                        required=True,
                        help="Critical Threshold")
    
    args = parser.parse_args()

    # Connect to MailStore Server Administration API
    api = mailstore.mgmt.ServerClient(username=args.username, password=args.password, host=args.host, port=args.port,
                                      ignoreInvalidSSLCerts=True)

    # Figure out how many licenses are currently in use and available
    try:
        usedLicenses = len(api.GetUsers()["result"]) - 1    # -1 because admin does not count
        availableLicenses = args.licensed - usedLicenses
    except Exception:
        print("UNKNOWN - Cannot retrieve licensing information")
        sys.exit(3)

    # Finally determine the status to return
    if availableLicenses <= args.critical:
        print("CRITICAL - Number of available licenses ({2}) is below critical threshold ({3})|"
              "Licensed={0}, Used={1}, Available={2}".format(args.licensed, usedLicenses,
                                                             availableLicenses, args.critical))
        sys.exit(2)
    elif availableLicenses <= args.warning:
        print("WARNING - Number of available licenses ({2}) is below warning threshold ({3})|"
              "Licensed={0}, Used={1}, Available={2}".format(args.licensed, usedLicenses,
                                                             availableLicenses, args.warning))
        sys.exit(1)
    else:
        print("OK - Number of available licenses: {2}|"
              "Licensed={0}, Used={1}, Available={2}".format(args.licensed, usedLicenses, availableLicenses))
        sys.exit(0)
